#pragma once


#include<dune/pdelab/constraints/common/constraints.hh>
#include<dune/pdelab/constraints/conforming.hh>

#include<dune/pdelab/gridfunctionspace/gridfunctionspace.hh>
#include<dune/pdelab/gridfunctionspace/subspace.hh>
#include <dune/pdelab/gridfunctionspace/vectorgridfunctionspace.hh>

//_________________________________________________________________________________________________________________________


template<typename RF, typename ES, typename V_FEM, typename P_FEM, typename PHI_FEM, typename U_FEM>
class GFS_fs_chns_r
{
public:
  typedef Dune::PDELab::ConformingDirichletConstraints ConstraintsAssembler;
  typedef Dune::PDELab::ISTL::VectorBackend<> VectorBackend;
  typedef Dune::PDELab::ISTL::VectorBackend<Dune::PDELab::ISTL::Blocking::fixed> VB_Block;
  typedef Dune::PDELab::LexicographicOrderingTag OrderingTagL;
  typedef Dune::PDELab::EntityBlockedOrderingTag OrderingTagB;

  ConstraintsAssembler conp, conphi, conmu, conu;

  typedef Dune::PDELab::VectorGridFunctionSpace<ES,V_FEM,ES::GridView::dimension,
                                                VectorBackend,VectorBackend,ConstraintsAssembler> V_GFS;
  V_GFS vGfs;

  typedef Dune::PDELab::GridFunctionSpace<ES, P_FEM, ConstraintsAssembler, VectorBackend> P_GFS;
  P_GFS pGfs;

  typedef Dune::PDELab::GridFunctionSpace<ES, PHI_FEM, ConstraintsAssembler, VectorBackend> PHI_GFS;
  PHI_GFS phiGfs;

  typedef Dune::PDELab::GridFunctionSpace<ES, PHI_FEM, ConstraintsAssembler, VectorBackend> MU_GFS;
  MU_GFS muGfs;

  typedef Dune::PDELab::GridFunctionSpace<ES, U_FEM, ConstraintsAssembler, VectorBackend> U_GFS;
  U_GFS uGfs;

  typedef Dune::PDELab::CompositeGridFunctionSpace<VectorBackend, OrderingTagL, V_GFS, P_GFS> NS;
  NS ns;

  typedef Dune::PDELab::CompositeGridFunctionSpace<VB_Block, OrderingTagB, PHI_GFS, MU_GFS, U_GFS> CH;
  CH ch;

  typedef typename NS::template ConstraintsContainer<RF>::Type NS_CC;
  NS_CC nsCC;

  typedef typename CH::template ConstraintsContainer<RF>::Type CH_CC;
  CH_CC chCC;

  GFS_fs_chns_r(const ES& es, const V_FEM& vFem, const P_FEM& pFem, const PHI_FEM& phiFem, const U_FEM& uFem) :
      conp(),conphi(),conmu(),
      vGfs(es,vFem),
      pGfs(es,pFem, conp),
      phiGfs(es,phiFem, conphi),
      muGfs(es,phiFem, conmu),
      uGfs(es,uFem,conu),
      ns(vGfs, pGfs),
      ch(phiGfs, muGfs, uGfs),
      nsCC(),
      chCC()
  {
    vGfs.name("velocity");
    pGfs.name("pressure");
    phiGfs.name("phasefield");
    muGfs.name("potential");
    uGfs.name("concentration");
  }

  template<typename Bdry>
  void updateConstraintsContainer(const Bdry& bdry)
  {
    Dune::PDELab::constraints(bdry.bdryNS,ns,nsCC, 0);
    Dune::PDELab::constraints(bdry.bdryCH,ch,chCC, 0);
  }
};

//_________________________________________________________________________________________________________________________


template<typename GFS, typename NS_V, typename CH_V>
class DGF_fs_chns_r
{
public:

  typedef Dune::PDELab::GridFunctionSubSpace<typename GFS::NS,Dune::TypeTree::TreePath<0> > VSubGFS;
  VSubGFS vSubGfs;
  typedef Dune::PDELab::VectorDiscreteGridFunction<VSubGFS,NS_V> VDGF;
  VDGF vDgf;

  typedef Dune::PDELab::GridFunctionSubSpace<typename GFS::NS,Dune::TypeTree::TreePath<1> > PSubGFS;
  PSubGFS pSubGfs;
  typedef Dune::PDELab::DiscreteGridFunction<PSubGFS,NS_V> PDGF;
  PDGF pDgf;

  typedef Dune::PDELab::GridFunctionSubSpace<typename GFS::CH,Dune::TypeTree::TreePath<0> > PhiSubGFS;
  PhiSubGFS phiSubGfs;
  typedef Dune::PDELab::DiscreteGridFunction<PhiSubGFS,CH_V> PhiDGF;
  PhiDGF phiDgf;

  typedef Dune::PDELab::GridFunctionSubSpace<typename GFS::CH,Dune::TypeTree::TreePath<1> > MuSubGFS;
  MuSubGFS muSubGfs;
  typedef Dune::PDELab::DiscreteGridFunction<MuSubGFS,CH_V> MuDGF;
  MuDGF muDgf;

  typedef Dune::PDELab::GridFunctionSubSpace<typename GFS::CH,Dune::TypeTree::TreePath<2> > USubGFS;
  USubGFS uSubGfs;
  typedef Dune::PDELab::DiscreteGridFunction<USubGFS,CH_V> UDGF;
  UDGF uDgf;


  DGF_fs_chns_r(const GFS& gfs, const NS_V& nsV, const CH_V& chV) :
      vSubGfs(gfs.ns),
      vDgf(vSubGfs,nsV),
      pSubGfs(gfs.ns),
      pDgf(pSubGfs,nsV),
      phiSubGfs(gfs.ch),
      phiDgf(phiSubGfs,chV),
      muSubGfs(gfs.ch),
      muDgf(muSubGfs,chV),
      uSubGfs(gfs.ch),
      uDgf(uSubGfs,chV)
  {
  }

  template<typename VTKWRITER>
  void addToVTKWriter(VTKWRITER& vtkwriter)
  {
    vtkwriter.addVertexData(std::make_shared<Dune::PDELab::VTKGridFunctionAdapter<VDGF> >(vDgf,"v"));
    vtkwriter.addVertexData(std::make_shared<Dune::PDELab::VTKGridFunctionAdapter<PDGF> >(pDgf,"p"));
    vtkwriter.addVertexData(std::make_shared<Dune::PDELab::VTKGridFunctionAdapter<PhiDGF> >(phiDgf,"phi"));
    vtkwriter.addVertexData(std::make_shared<Dune::PDELab::VTKGridFunctionAdapter<MuDGF> >(muDgf,"mu"));
    vtkwriter.addVertexData(std::make_shared<Dune::PDELab::VTKGridFunctionAdapter<UDGF> >(uDgf,"u"));
  }
};

//_________________________________________________________________________________________________________________________

template<typename IniV, typename IniP, typename IniPhi, typename IniMu, typename IniU>
class Ini_fs_chns_r
{
public:
  IniV iniV;
  IniP iniP;
  IniPhi iniPhi;
  IniMu iniMu;
  IniU iniU;
  typedef Dune::PDELab::CompositeGridFunction<IniV,IniP> IniNS;
  typedef Dune::PDELab::CompositeGridFunction<IniPhi,IniMu,IniU> IniCH;
  IniNS iniNS;
  IniCH iniCH;

  template<typename GV, typename Params>
  Ini_fs_chns_r(const GV& gv, const Params& params) :
      iniV(gv, params),
      iniP(gv, params),
      iniPhi(gv, params),
      iniMu(gv, params),
      iniU(gv, params),
      iniNS(iniV,iniP),
      iniCH(iniPhi,iniMu,iniU)
  {
  }
};

//_________________________________________________________________________________________________________________________

template<const int dim, typename BdryV_Scalar, typename BdryP, typename BdryPhi, typename BdryMu, typename BdryU>
class Bdry_fs_chns_r
{
public:
  BdryV_Scalar bdryV_Scalar;
  BdryP bdryP;
  BdryPhi bdryPhi;
  BdryMu bdryMu;
  BdryU bdryU;
  typedef Dune::PDELab::PowerConstraintsParameters<BdryV_Scalar,dim>    BdryV;
  typedef Dune::PDELab::CompositeConstraintsParameters<BdryV, BdryP> BdryNS;
  typedef Dune::PDELab::CompositeConstraintsParameters<BdryPhi, BdryMu, BdryU> BdryCH;
  BdryV bdryV;
  BdryNS bdryNS;
  BdryCH bdryCH;

  Bdry_fs_chns_r() :
      bdryV_Scalar(),
      bdryP(),
      bdryPhi(),
      bdryMu(),
      bdryU(),
      bdryV(bdryV_Scalar),
      bdryNS(bdryV,bdryP),
      bdryCH(bdryPhi,bdryMu,bdryU)
  {
  }
};

template<typename BdryV1, typename BdryV2, typename BdryP, typename BdryPhi, typename BdryMu, typename BdryU>
class Bdry_fs_chns_r_compositeV_2d
{
public:
  const int dim=2;

  BdryV1 bdryV1;
  BdryV2 bdryV2;
  BdryP bdryP;
  BdryPhi bdryPhi;
  BdryMu bdryMu;
  BdryU bdryU;
  typedef Dune::PDELab::CompositeConstraintsParameters<BdryV1,BdryV2>    BdryV;
  typedef Dune::PDELab::CompositeConstraintsParameters<BdryV, BdryP> BdryNS;
  typedef Dune::PDELab::CompositeConstraintsParameters<BdryPhi, BdryMu, BdryU> BdryCH;
  BdryV bdryV;
  BdryNS bdryNS;
  BdryCH bdryCH;

  Bdry_fs_chns_r_compositeV_2d() :
      bdryV1(),
      bdryV2(),
      bdryP(),
      bdryPhi(),
      bdryMu(),
      bdryU(),
      bdryV(bdryV1,bdryV2),
      bdryNS(bdryV,bdryP),
      bdryCH(bdryPhi,bdryMu,bdryU)
  {
  }
};

template<typename BdryV1, typename BdryV2, typename BdryP,
          typename BdryPhi, typename BdryMu, typename BdryU>
class Bdry_fs_chns_r_compositeV_2d_Explicit_Domain
{
public:
  BdryV1 bdryV1;
  BdryV2 bdryV2;
  BdryP bdryP;
  BdryPhi bdryPhi;
  BdryMu bdryMu;
  BdryU bdryU;
  typedef Dune::PDELab::CompositeConstraintsParameters<BdryV1,BdryV2>    BdryV;
  typedef Dune::PDELab::CompositeConstraintsParameters<BdryV, BdryP> BdryNS;
  typedef Dune::PDELab::CompositeConstraintsParameters<BdryPhi, BdryMu, BdryU> BdryCH;
  BdryV bdryV;
  BdryNS bdryNS;
  BdryCH bdryCH;

  template<typename DomainSize>
  Bdry_fs_chns_r_compositeV_2d_Explicit_Domain(DomainSize ds) :
      bdryV1(ds),
      bdryV2(ds),
      bdryP(ds),
      bdryPhi(ds),
      bdryMu(ds),
      bdryU(ds),
      bdryV(bdryV1,bdryV2),
      bdryNS(bdryV,bdryP),
      bdryCH(bdryPhi,bdryMu,bdryU)
  {
  }
};
